&time_control
run_days = 0,
run_hours = 6,
run_minutes = 0,
run_seconds = 0,
start_year = 2010,
start_month = 06,
start_day = 08,
start_hour = 18,
start_minute = 00,
start_second = 00,
end_year = 2010,
end_month = 06,
end_day = 09,
end_hour = 00,
end_minute = 00,
end_second = 00,
interval_seconds = 21600,
input_from_file = .true.,
history_interval = 60,
frames_per_outfile = 1,
restart = .false.,
restart_interval = 2881,
io_form_history = 2,
io_form_restart = 102,
io_form_input = 2,
io_form_boundary = 2,
io_form_auxhist2 = 2,
debug_level = 0
diag_print = 0
/

&domains
time_step = 60,
time_step_fract_num = 0,
time_step_fract_den = 1,
max_dom = 1,
e_we = 415,
e_sn = 325,
e_vert = 51,
num_metgrid_levels = 32
num_metgrid_soil_levels = 4
dx = 15000,
dy = 15000,
grid_id = 1,
parent_id = 0,
i_parent_start = 1,
j_parent_start = 1,
parent_grid_ratio = 1,
parent_time_step_ratio = 1,
feedback = 0,
smooth_option = 1,
p_top_requested = 1500,
lagrange_order = 2
interp_type = 2
interp_theta = .false.
hypsometric_opt = 2
extrap_type = 2
t_extrap_type = 2
use_surface = .true.
use_levels_below_ground = .true.
lowest_lev_from_sfc = .false.,
force_sfc_in_vinterp = 1
zap_close_levels = 500
eta_levels = 1.0000, 0.9980, 0.9940, 0.9870, 0.9750, 0.9590,
0.9390, 0.9160, 0.8920, 0.8650, 0.8350, 0.8020, 0.7660,
0.7270, 0.6850, 0.6400, 0.5920, 0.5420, 0.4970, 0.4565,
0.4205, 0.3877, 0.3582, 0.3317, 0.3078, 0.2863, 0.2670,
0.2496, 0.2329, 0.2188, 0.2047, 0.1906, 0.1765, 0.1624,
0.1483, 0.1342, 0.1201, 0.1060, 0.0919, 0.0778, 0.0657,
0.0568, 0.0486, 0.0409, 0.0337, 0.0271, 0.0209, 0.0151,
0.0097, 0.0047, 0.0000
/

&physics
mp_physics = 8,
ra_lw_physics = 4,
ra_sw_physics = 4,
radt = 10,
sf_sfclay_physics = 2,
sf_surface_physics = 2,
bl_pbl_physics = 2,
bldt = 0,
cu_physics = 6,
cudt = 5,
mp_zero_out = 2,
mp_zero_out_thresh = 1.e-10,
isfflx = 1,
ifsnow = 0,
icloud = 1,
surface_input_source = 1,
num_land_cat = 20,
num_soil_layers = 4,
o3input = 2,
aer_opt = 1,
sst_skin = 1,
/

&dynamics
w_damping = 1,
diff_opt = 1,
km_opt = 4,
damp_opt = 3,
zdamp = 5000., 5000., 5000.,
dampcoef = 0.2, 0.01, 0.2,
non_hydrostatic = .true., .true.,
moist_adv_opt = 2, 2,
scalar_adv_opt = 2, 2,
diff_6th_opt = 2, 2,
diff_6th_factor = 0.25, 0.12,
epssm = 0.1
/

&bdy_control
spec_bdy_width = 5,
spec_zone = 1,
relax_zone = 4,
specified = .true., .false.,
nested = .false.,.true.,
/

&namelist_quilt
nio_tasks_per_group = 2,
nio_groups = 1,
/

&wrfvar1
check_max_iv_print = .false.,
write_increments = .false.,
/

&wrfvar2
/

&wrfvar3
/

&wrfvar4
use_synopobs = .false.,
use_shipsobs = .false.,
use_metarobs = .false.,
use_soundobs = .false.,
use_pilotobs = .false.,
use_airepobs = .false.,
use_geoamvobs = .false.,
use_polaramvobs = .false.,
use_bogusobs = .false.,
use_buoyobs = .false.,
use_profilerobs = .false.,
use_satemobs = .false.,
use_gpspwobs = .false.,
use_gpsrefobs = .false.,
use_qscatobs = .false.,
use_radarobs = .false.,
use_radar_rv = .false.,
use_radar_rf = .false.,
use_airsretobs = .false.,
/

&wrfvar5
check_max_iv = .false.,
put_rand_seed = .true.,
/

&wrfvar6
max_ext_its = 1
ntmax = 200,
eps = 0.01,
/

&wrfvar7
cv_options = 3,
as1 = PERT_SCALING, 2.0, 1.5,
as2 = PERT_SCALING, 2.0, 1.5,
as3 = PERT_SCALING, 2.0, 1.5,
as4 = PERT_SCALING, 2.0, 1.5,
as5 = PERT_SCALING, 2.0, 1.5,
rf_passes = 6,
var_scaling1 = 1.0,
var_scaling2 = 1.0,
var_scaling3 = 1.0,
var_scaling4 = 1.0,
var_scaling5 = 1.0,
len_scaling1 = 1.0,
len_scaling2 = 1.0,
len_scaling3 = 1.0,
len_scaling4 = 1.0,
len_scaling5 = 1.0,
je_factor = 1.0,
/

&wrfvar8
/

&wrfvar9
trace_use = .false.,
trace_memory = .false.,
trace_all_pes = .false.,
trace_csv = .false.,
use_html = .false.,
/
&wrfvar10
/

&wrfvar11
cv_options_hum = 1,
check_rh = 1,
set_omb_rand_fac = 1.0,
seed_array1 = 1,
seed_array2 = 1,
/

&wrfvar12
/

&wrfvar13
vert_corr = 2,
vertical_ip = 0,
vert_evalue = 1,
max_vert_var1 = 99.0,
max_vert_var2 = 99.0,
max_vert_var3 = 99.0,
max_vert_var4 = 99.0,
max_vert_var5 = 0.0,
/

&wrfvar14
/

&wrfvar15
num_pseudo = 0,
/

&wrfvar16
/

&wrfvar17
analysis_type = 'RANDOMCV',
/

&wrfvar18
analysis_date = 'yyyy-mm-dd_hh:mm:ss.0000',
/

&wrfvar19
/

&wrfvar20
/

&wrfvar21
/

&wrfvar22
/

&wrfvar23
/