// Initialize Mermaid when the page loads
document.addEventListener('DOMContentLoaded', function() {
  mermaid.initialize({
    startOnLoad: true,
    theme: "default",
    flowchart: {
      htmlLabels: false,
      useMaxWidth: false
    },
    er: {
      useMaxWidth: false
    },
    sequence: {
      useMaxWidth: false,
      noteFontWeight: "14px",
      actorFontSize: "14px",
      messageFontSize: "16px"
    },
    journey: {
      useMaxWidth: false
    },
    gitGraph: {
      useMaxWidth: false
    }
  });
});